#!/usr/bin/env python
# encoding: utf-8
"""
This module contains code for the use of finite difference stencils as
differential operators.

Stencil values are calculated using Fornberg's paper. Code for the
calculations is available in
`finite_difference_weights_generator.py` located in Code/Utils.
See Fornberg's paper for the notation.
"""

from builtins import object
from coffee.settings import be

################################################################################
# Finite difference default ghost point processor
################################################################################
def ghost_point_processor(data, b_values):
    """A utility function that understands how data defined over ghost points
    for finite difference stencils should be handled when communicated via MPI."""
    for b_slice, b_data in b_values:
        data[b_slice] = b_data

################################################################################
# Finite difference stencil base class.
################################################################################


class FD_stencil(object):
    """This class takes an array of numbers and uses them to act on a
    vector of numbers.

    That is an FD_stencil represents
    a 'standard' finite difference stencil, e.g.:

        [-1, 2, -1]

    as a linear operator.
    """

    def __init__(self, stencil_coefs, loffset, roffset):
        """The initialiser for an FD_stencil.

        Parameters
        ==========
        stencil_coefs : list of floats
            Gives a finite difference stencil over a fixed step size.
        loffset : int
            How many elements of the `stencil_coefs` occurs before the
            point at which the derivative is being calculated.
        roffset: int
            Similar to loffset, but counted from the right.
        """
        self.stencil_coefs = stencil_coefs
        self.loffset = loffset
        self.roffset = roffset

    def __repr__(self):
        return "<FD_stencil %s, [%d,%d]>" % (
            repr(self.stencil_coefs),
            self.loffset,
            self.roffset,
        )

    def __call__(self, u, apply_at=None):
        """Applies the finite difference stencil to calculate a vector of
        derivatives of the function represented by the given data.

        Parameters
        ==========
        u : tslice.TimeSlice
            The data to be differentiated.
        apply_at : int, Optional
            The index of the data at which the derivative is to be calculated.
            If not given then all derivatives are calculated.

        Returns
        =======
        numpy.ndarray:
            A list of derivatives, or a list with one derivative.
        """
        if apply_at is not None:
            lbound = apply_at - self.loffset
            rbound = apply_at + self.roffset + 1
            if rbound == 0:
                rbound = None
            return be.dot(self.stencil_coefs, u[lbound:rbound])
        else:
            """The convolve method does not quite do the right thing, it
            flips the direction of iteration. Hence the ::-1 below.

            Note also that the convolve method is based on the
            multiarray.correlate a C routine."""
            return be.convolve(u, self.stencil_coefs[::-1], mode="same")


def flip(stencil):
    """This method takes, for example, a forward difference operator and returns
    a backwards difference operator using 'the same' stencil.

    For example:
        flip(F12_stencil) = B12_stencil
        flip(C12_stencil) = C12_stencil

    Parameters
    ==========
    stencil : list of floats
        A 'standard' finte difference stencil over a fixed grid step.
    """
    return FD_stencil(-stencil.stencil_coefs[::-1], stencil.roffset, stencil.loffset)


################################################################################
# First derivative stencils
################################################################################
"""Further stencil can be calculated using 
`finite_difference_weights_generator.py' located in Code/Utils.
I suggest reading Fornberg's paper for the notation."""


class C12_stencil(FD_stencil):
    """The standard central second order first derivative finite difference."""

    def __init__(self):
        stencil = be.array([-1.0 / 2.0, 0.0, 1.0 / 2.0])
        loff = 1
        roff = 1
        super(C12_stencil, self).__init__(stencil, loff, roff)


class C14_stencil(FD_stencil):
    """The standard central fourth order first derivative finite difference."""

    def __init__(self):
        stencil = be.array([1.0 / 12, -2.0 / 3, 0, 2.0 / 3, -1.0 / 12])
        loff = 2
        roff = 2
        super(C14_stencil, self).__init__(stencil, loff, roff)


class F12_stencil(FD_stencil):
    """The standard forward second order first derivative finite difference."""

    def __init__(self):
        stencil = be.array([-3.0 / 2.0, 2, -1.0 / 2.0])
        loff = 0
        roff = 2
        super(F12_stencil, self).__init__(stencil, loff, roff)


class F13_stencil(FD_stencil):
    """The standard forward third order first derivative finite difference."""

    def __init__(self):
        stencil = be.array([-11.0 / 6.0, 3.0, -3.0 / 2.0, 1.0 / 3])
        loff = 0
        roff = 3
        super(F13_stencil, self).__init__(stencil, loff, roff)


class F14_stencil(FD_stencil):
    """The standard forward fourth order first derivative finite difference."""

    def __init__(self):
        stencil = be.array([-25.0 / 12, 4.0, -3.0, 4.0 / 3, -1.0 / 4])
        loff = 0
        roff = 4
        super(F14_stencil, self).__init__(stencil, loff, roff)


class MF14_stencil(FD_stencil):
    """An offset forward fourth order first derivative finite difference
    that calculates the derivative at the second grid point."""

    def __init__(self):
        stencil = be.array([-1.0 / 4, -15.0 / 18, 3.0 / 2, -1.0 / 2, 15.0 / 180])
        loff = 1
        roff = 3
        super(MF14_stencil, self).__init__(stencil, loff, roff)


class B12_stencil(FD_stencil):
    """The standard backward second order first derivative finite difference."""

    def __init__(self):
        stencil = be.array([1.0 / 2.0, -2.0, 3.0 / 2.0])
        loff = 2
        roff = 0
        super(B12_stencil, self).__init__(stencil, loff, roff)


################################################################################
# Second derivative stencils
################################################################################


class C22_stencil(FD_stencil):
    """The standard central second order second derivative finite difference."""

    def __init__(self):
        stencil = be.array([1.0, -2.0, 1.0])
        loff = 1
        roff = 1
        super(C22_stencil, self).__init__(stencil, loff, roff)


class F22_stencil(FD_stencil):
    """The standard forward second order second derivative finite difference."""

    def __init__(self):
        stencil = be.array([2.0, -5.0, 4.0, -1.0])
        loff = 0
        roff = 3
        super(F22_stencil, self).__init__(stencil, loff, roff)


class B22_stencil(FD_stencil):
    """The standard backward second order second derivative finite difference."""

    def __init__(self):
        stencil = be.array([1.0, -4.0, 5.0, -2.0])
        loff = 3
        roff = 0
        super(B22_stencil, self).__init__(stencil, loff, roff)


################################################################################
# Finite Difference base class
################################################################################


class FD_diffop(object):
    """The FD_diffop class represents the action of a collection of
    FD_stencils on a vector of values."""

    name = "FD_diffop"

    def __init__(self):
        pass

    def __call__(self, u, dx):
        """Gives the result of the linear operator represented by this
        class on a vector of data.

        Parameters
        ==========
        u : tslice.TimeSlice
            The data to take a derivative of.
        dx: float
            The spatial step size.

        Returns
        =======
        numpy.ndarray:
            The vector of derivatives.
        """
        ru = self.central(u)
        for i, b in self.boundaries:
            ru[i] = b(u, apply_at=i)
        return ru / (dx**self.order)

    def __str__(self):
        return "Differential operator " % self.name

    def save(self):
        """This is a convenience method that saves a textual representation,
        using be.savetxt, of the data used in the finite difference operator
        to a file in the users home directory."""
        filename = os.path.expanduser("~/" + self.name)
        print(filename)
        be.savetxt(filename + "_left.txt", self.central)
        be.savetxt(filename + "_right.txt", self.boundaries)

    def ghost_points(self):
        """Returns the number of ghost points for this operator.

        See the mpi module for documentation of ghost points.

        Returns
        =======
        two tuple of ints:
            The number of ghost points for the left and right application of
            this operator.
        """
        return self.central.loffset, self.central.roffset

    def internal_points(self):
        """Returns the number of internal points for this operator.

        See the mpi module for documentation of internal points.

        Returns
        =======
        two tuple of ints:
            The number of interal points for the left and right application of
            this operator.
        """
        return self.ghost_points()


################################################################################
# First Derivative operators
################################################################################
class FD12(FD_diffop):
    """Implements a second order FD routine for first order derivatives.

    On the boundaries uses forward/backward differences. Sufficiently
    away from the boundaries uses a central scheme. Assumes that the
    FD grid is evenly spaced."""

    name = "FD12"
    central = C12_stencil()
    boundaries = be.array([(0, F12_stencil()), (-1, B12_stencil())])
    order = 1


class FD14(FD_diffop):
    """Implements a fourth order FD routine for first order derivatives.

    On the boundaries uses forward/backward differences. On
    interor points where the central difference scheme cannot be applied,
    skewed forward/backward difference stencils are used. Sufficiently
    away from the boundaries uses a central scheme. Assumes that the
    FD grid is evenly spaced."""

    name = "FD14"
    central = C14_stencil()
    boundaries = be.array(
        [
            (0, F14_stencil()),
            (1, MF14_stencil()),
            (-2, flip(MF14_stencil())),
            (-1, flip(F14_stencil())),
        ]
    )
    order = 1


################################################################################
# Second Derivative operators
################################################################################


class FD22(FD_diffop):
    """Implements a second order FD routine for second order derivatives.

    On the boundaries uses forward/backward differences. Sufficiently
    away from the boundaries uses a central scheme. Assumes that the
    FD grid is evenly spaced."""

    name = "FD22"
    central = C22_stencil()
    boundaries = be.array([(0, F22_stencil()), (-1, B22_stencil())])
    order = 2



#!/usr/bin/env python
# encoding: utf-8


"""
OneDAdvection_plotter.py

Created by Chris Stevens 2023
"""

# Import Python libraries
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

from coffee.actions import Prototype

class Plotter(Prototype):
    def __init__(self, system, frequency = 1, xlim = (-1,1), ylim = (-1,1), \
                 delay = 0.0):
        super(Plotter, self).__init__(frequency)
        self.system = system
        self.delay  = delay
        plt.ion()
        fig = plt.figure(0,figsize=(10,10))
        ax = fig.add_subplot(111)
        self.line = ax.add_line(Line2D(xlim, ylim, label=r"$\Psi$"))
        
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)       
        ax.grid(True)
        self.axes = ax
        fig.show()

    def _doit(self, it, U):
        x   = U.domain.meshes[0]
        Psi = U.data[0]
        
        self.axes.set_title("Iteration: %d, Time: %f" % (it,U.time))
        plt.ioff()

        self.line.set_xdata(x)
        self.line.set_ydata(Psi)
        self.line.set_color('b')
       
        self.axes.legend(loc = 'lower center',prop={'size': 24},ncol=3)

        plt.ion()
        plt.figure(0)
        plt.draw()
        plt.pause(self.delay)


#!/usr/bin/env python
# encoding: utf-8
"""
A timeslice is the package which contains all data necessary to represent a
solution of the system at a given point in time.

This module contains the abstract base class (abc) for TimeSlice objects and a
default implementation that makes certain assumptions. It is highly recommended
that you subclass either of these classes.

As usual the abc is intended to make you think twice about doing things
differently. All other classes that interact with TimeSlice objects assume that
the interface is as given below.

Now it should be noted that the variables ABCTimeSlice.data,
ABCTimeSlice.domain and ABCTimeSlice.time should have been declared as
abc.abstractproperty. Doing this is possible, but because of a limitation in
the abc class for Python 2 a substantial amount of additional code is needed.
So this hasn't been done. In fact NONE of the methods or variables are declared
as abstract. 

What is recommended is that you subclass ABCTimeSlice and call it's functions
once appropriate checks have been carried out.

At some point this will need to be changed... I assume after the move to Python
3 is made.

TimeSlice objects should contain all information needed to interperet the
values of the functions being numerically evolved. In particular, aim to ensure
that you can restart a simulation if given a timeslice, system and solver.
"""

from builtins import object
import abc
from coffee.settings import be

###############################################################################
# TimeSlice Abstract Base Class
##############################################################################
class ABCTimeSlice(object, metaclass=abc.ABCMeta):
    """The abstract base class for TimeSlice objects.

    The interface below is assumed by all other classes that interact with
    TimeSlice objects (which is just about everything).

    """

    def __init__(self, data, domain, time, name=None, *args, **kwds):
        """Make a TimeSlice instance!

        It is recommended to call this method once your subclass has done
        appropriate vetting of data, domain and time. Not only because this
        ensures things are stored in the right places, but also because it sets
        up a logging object, logging.getLogger(self.name).

        Parameters
        ----------

        data :
            The values of the functions being solved for.
        domain :
            The grid over which the values are calcualted.
        time : float
            The time for which the values in data are valid.

        name : string
            The name of your subclass. This defaults to ABCTimeSlice and is
            used during logging.

        """
        self.data = data
        self.domain = domain
        self.time = time
        if name is None:
            self.name = "ABCTimeSlice"
        else:
            self.name = name
        super(ABCTimeSlice, self).__init__(*args, **kwds)

    def __repr__(self):
        """Returns a naive string representation of the slice.

        Returns
        -------
        string

        """
        s = "%s(data = %s, domain = %s, time = %s)" % (
            self.name,
            repr(self.data),
            repr(self.domain),
            repr(self.time),
        )
        return s

    def external_slices(self):
        """Returns a list of tuples that describe boundaries of grids for
        external boundaries.

        This is purely for convenience so that users don't need to
        write tslice.domain.external_slices().

        Returns
        -------
        list, slices
            The result of a call to self.domain.external_slices(self.data.shape)

        """
        return self.domain.external_slices(self.data.shape)

    def communicate(self, ghost_point_processor=None, data=None):
        """Returns a list of tuples that describe the boundaries of
        grids for internal boundaries.

        This is currently used when the full domain is divided into
        subdomains for use with mpi.
        This is purely for convenience so that users don't need to
        write tslice.domain.communicate().

        Parameters
        ----------
        ghost_point_processor : function
            A function with signature (data, list of slices). See the
            documentation for ABCGrid.communicate() for more information.

        data :
            The information to be communicated. It defaults to self.data.

        Returns
        -------
        list, slices
            The result of a call to self.domain.communicate(self.data)

        """
        return self.domain.communicate(
            data if data is not None else self.data,
            ghost_point_processor=ghost_point_processor,
        )

    def barrier(self):
        """A wrapper to self.domain.barrier()."""
        return self.domain.barrier()

    def collect_data(self):
        """A wrapper to self.domain.collect_data.

        Returns a timeslice containing the almalgum of all data and domains
        across all subdomains. It allocates new memory for the returned
        timeslice.
        This is currently used to pass non-distributed data to actions in the
        ibvp method.

        Returns
        -------
        tslice.TimeSlice
            A single timeslice which represents the complete data and
            domain for that point of time.

        """
        data_all = self.domain.collect_data(self.data)
        domain_all = self.domain.full_grid
        r_tslice = self.__class__(data_all, domain_all, self.time)
        return r_tslice

    def __add__(self, other):
        # It is very important that the rv is other + self.data not
        # self.data + other.
        # The reason (seems) is because as self.data can be an
        # nd.array the sum self.data + other
        # ends up being computed as the elements of self.other
        # plus other, itself an nd.array.
        # This screws up the array of elements.
        # Putting other first ensures that if other is a timeslice
        # then the sum isn't distributed over the elements of self.data.
        try:
            rv = other + self.data
        except:
            raise NotImplementedError(
                "Addition of %s and %s is not implemented" % (self, other)
            )
        if isinstance(rv, self.__class__):
            return rv
        else:
            return self.__class__(rv, self.domain, self.time, name=self.name)

    def __iadd__(self, other):
        if isinstance(other, self.__class__):
            self.data += other.data
        else:
            try:
                self.data += other
            except:
                raise NotImplementedError(
                    "Addition of %s and %s is not implemented" % (self, other)
                )
        return self

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        try:
            rv = self.data - other
        except:
            raise NotImplementedError(
                "Subtraction of %s and %s is not implemented" % (self, other)
            )
        return self.__class__(rv, self.domain, self.time)

    def __isub__(self, other):
        try:
            self.data -= other
        except:
            raise NotImplementedError(
                "In place subraction of %s and %s is not implemented" % (self, other)
            )
        return self

    def __rsub__(self, other):
        try:
            r_time_slice = other - self.data
        except:
            raise NotImplementedError(
                "Reflected subtraction of %s and %s is not implemented" % (other, self)
            )
        return r_time_slice

    def __mul__(self, other):
        try:
            rv = self.data * other
        except:
            raise NotImplementedError(
                "Multiplication of %s and %s is not implemented" % (self, other)
            )
        return self.__class__(rv, self.domain, self.time)

    def __imul__(self, other):
        try:
            self.data *= other
        except:
            raise NotImplementedError(
                "In place multiplicatio of %s and %s is not implemented" % (self, other)
            )
        return self

    def __rmul__(self, other):
        return self * other

    def __div__(self, other):
        try:
            rv = self.data / other
        except:
            raise NotImplementedError(
                "Division of %s and %s is not implemented" % (self, other)
            )
        return self.__class__(rv, self.domain, self.time)

    def __idiv__(self, other):
        try:
            self.data /= other
        except:
            raise NotImplementedError(
                "In place division of %s and %s is not implemented" % (self, other)
            )
        return self

    def __truediv__(self, other):
        return self.__div__(other)

    def __itruediv__(self, other):
        return self.idiv(other)

    def __pow__(self, other):
        try:
            rv = self.data**other
        except:
            raise NotImplementedError(
                "Exponentiation of %s by %s is not implemented" % (self, other)
            )
        return self.__class__(rv, self.domain, self.time)

    def __ipow__(self, other):
        try:
            self.data **= other
        except:
            raise NotImplementedError(
                "In place exponentiation of %s by %s is not implemented" % (self, other)
            )
        return self


###############################################################################
# Concrete implementations
###############################################################################
class TimeSlice(ABCTimeSlice):
    """A default subclass of ABCTimeSlice.

    This implementation assumes that all data are numpy arrays of the same
    shape. If your data is not like this then you should subclass ABCTimeSlice.
    This might be appropriate when working with spectral methods, for example.

    Due to the assumption that the data is represented as a numpy array, this
    class also implements a large number of methods which allow this object to
    be added, multiplied, etc...

    To be honest rather than implementing all the additional methods by hand
    it'd be easier just to make this default TimeSlice a subclass of be.ndarray
    itself and allow TimeSlice.data to access the underlying array. But this
    hasn't been done.
    """

    def __init__(self, data, *args, **kwds):
        """Instantiates a TimeSlice object.

        Calls the ABCTimeSlice __init__ method after handling the data.
        Thus you should refer to the documentation for ABCTimeSlice.__init__()
        for details of other parameters.

        Parameters
        ----------
        data :
            Converts data to a numpy array before passing it on to ABCTimeSlice.

        """
        data = be.array(data)
        if "name" not in kwds:
            kwds["name"] = "TimeSlice"
        super(TimeSlice, self).__init__(data, *args, **kwds)



#!/usr/bin/env python
# encoding: utf-8

"""
Grid objects represent the computational, spatial, domain.

Each grid object understands it's own discretisation as well as how
"""


from builtins import range
from builtins import object
import abc

try:
    from coffee.mpi import mpiinterfaces
except ModuleNotFoundError:
    pass

from coffee.settings import be

################################################################################
# Base Boundary data class
################################################################################
class ABCBoundary(object, metaclass=abc.ABCMeta):
    """The abstract base class for boundary classes.

    A boundary class manages data associated to ghost points and the slices
    specifying which data points should be communicated between MPI nodes.

    Suppose that our computational domain is a line, represented as the
    array::

        domain = be.linspace(0,1,num=11)

    This grid has two external boundaries at indices 0 and -1.
    A subclass of ABCBoundary will generate a list of slices::

        [slice(0,None,None)

    that when applied to ``domain`` extract the values of the boundary points.

    Suppose that our computational domain is a line now represented as two
    sub-domains::

        domain_1 = be.linspace(0,1,num=11)[:6]
        domain_2 = be.linspace(0,1,num=11)[5:]

    Each array now has an external boundary, which represents a boundary of
    the global computational domain, and an internal boundary, which
    represents an artifical boundary generated by the representation of the
    global domain as two different arrays.

    A subclass of ABCBoundary should draw a distinction between internal
    boundaries and external boundaries. This is so that system objects
    can handle each type of boundary differently. The design goal
    is for the lists to be iterated over in the system object so that
    each boundary of each type can be appropriately treated.
    """

    # I use -1 and 1 to reduce friction with the mpi direction variable
    LEFT = -1
    RIGHT = 1

    DIRECTION_ERROR = ValueError("Direction must be +/- 1")

    @staticmethod
    def _direction_to_index(direction):
        """A utility that translates between the MPI direction variable and
        the index of a hypothetical
        numpy array that represents the data over a grid.

        This method assumes that the grid points increase in value with
        increasing index. Hence a direction of -1 (meaning left) returns 0
        the first element in the array and a direction of 1 (meaning right)
        returns -1 meaning the last element of the array.

        The method is used to build the list of slices that extract the
        relavant sub-array of data on the 'boundary' of the grid.

        Parameters
        ==========
        direction: int
            Can be either -1 or 1 in reference to the MPI direction.

        Returns
        =======
        int
            The index of a hypothetical array that contains the data in the
            'direction' end of the array. So, if the grid points inc

        Raises
        ======
        ValueError
            If direction is not -1 or 1.

        """
        if direction == -1:
            return 0
        elif direction == 1:
            return 1
        raise DIRECTION_ERROR

    @staticmethod
    def _empty_slice(dimensions):
        """A utilty to construct a list of empty slices with length dimensions.

        Parameters
        ==========
        dimensions: int
            The number of dimensions of the spatial domain of the function to
            be computed. Also: the number of dimensions in the grid.

        Returns
        =======
        list, slices
            Returns a list of empty slices of length dimension
        """
        return [slice(None, None, None) for i in range(dimensions)]

    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def ghost_points(self, dimension, direction):
        """Returns the number of ghost points for the specified dimension and
        direction.

        Parameters
        ==========
        dimension: int
            Which dimension of the domain are we asking for the number of
            ghost points?

        direction: int
            Of the one dimensional array representing the given dimension, which
            end of the array are we interested in knowing about the ghost points?

        Returns
        =======
        int
            The number of ghost points for the specified boundary.
        """

        return 0

    @abc.abstractmethod
    def internal_slice(self, shape, dimension, direction):
        """Returns a tuple of slices which, when applied
        to a data array, gives the data to be communicated to
        neighbouring grids.

        Parameters
        ==========
        shape: tuple of ints
            The shape of the array to be communicated. For example the shape
            of a numpy array representing the computational domain.

        dimension: int
            Which dimension do we need the list of slices for?

        direction: int
            Which direction do we need the slice for?

        Returns
        =======
        tuple of slices
            A list containing the slices that extract the data on the boundary.
        """
        return self._empty_slice(shape)

    @abc.abstractmethod
    def external_slice(self, shape, dimension, direction):
        """
        Returns a tuple which, when applied
        to a data array, gives the data on the extenal boundaries of the grid.

        Parameters
        ==========
        shape: tuple of ints
            The shape of the array to be communicated. For example the shape
            of a numpy array representing the computational domain.

        dimension: int
            Which dimension do we need the list of slices for?

        direction: int
            Which direction do we need the slice for?

        Returns
        =======
        list of slices
            A list containing the slices that extract the data on the boundary.
        """
        slices = self._empty_slice(len(shape))

        # +1 because for some reason I decided that vector data should be in the
        # first dimension.
        dim_index = dimension + 1
        if direction == -1:
            slices[dim_index] = slice(None, 1, None)
            return slices

        if direction == 1:
            slices[dim_index] = slice(-1, None, None)
            return slices

        raise DIRECTION_ERROR

    def external_slices(self, shape):
        """Returns a list of tuples. Each tuple contains an integer,
        representing the dimension, a direction and the result of
        calling external_slice(shape, dimension, direction).

        This is a convenience method to make iteration over external
        boundaries easy for the user.

        Parameters
        ==========
        shape: tuple of ints
            The shape of the computational domain. For example if the computational
            domain is a numpy array then the shape is the numpy shape of the array.

        Returns
        =======
        list of tuples of slices
            Each tuple of slices represents the sub array of the computational
            domain that made up of external boundary points.
        """

        # This implicitly assumes that the number of dimensions handled by
        # the grid object is the same as than managed by the mpi_comm
        dims = self.mpi_comm.Get_dim()
        neg_slices = [
            (i, -1, self.external_slice(shape, i, -1))
            for i in range(dims)
            if self.external_slice(shape, i, -1) != self._empty_slice(len(shape))
        ]
        pos_slices = [
            (i, 1, self.external_slice(shape, i, 1))
            for i in range(dims)
            if self.external_slice(shape, i, 1) != self._empty_slice(len(shape))
        ]
        return pos_slices + neg_slices


################################################################################
# Base Boundary data class
################################################################################
class SingleCartesianGridBoundary(ABCBoundary):
    """Implements ABCBoundary for a simulation on a single grid with no
    mpi dependence.

    That is a grid with no internal boundaries and every grid "edge" an
    external boundary. There are no ghost points."""

    pass


class MPIBoundary(ABCBoundary):
    """MPIBoundary implements the ABCBoundary class, it assumes use of an underlying
    MPI.Cartcomm object wrapped by mpi4py.

    The numbers of ghost points
    and "internal points" can be specified directly using arrays of two tuples.
    If the grid is two dimensional the, for example, the array of ghost points
    and internal points will look like::

        [(a, b), (c, d)]

    The tuple ``(a, b)`` gives the ghost / internal points for the first dimension.
    The int ``a`` gives the number of ghost / internal points for the
    left direction and ``b`` gives the number for the right direction.
    Similarly for ``(c,d)``.


    """

    def __init__(self, ghost_points, internal_points, mpi_comm=None, *args, **kwargs):
        """
        The initialiser for MPIBoundary.

        The ghost points and internal points variables must both be an array
        or tuple with one entry for each dimension. Each entry is a 2-tuple giving
        the number of points in the negative and positive directions for
        that axis. See class documentation for more information.

        Parameters
        ==========
        ghost_points: list of tuples of ints
            The number of ghost points in each direction for each dimension
        internal_points: list of tuples of ints
            The number of internal points in each direction for each dimension
        mpi_comm: MPI.Comm
            The mpi4py wrapper of MPI_COMM.

        """
        super(MPIBoundary, self).__init__(*args, **kwargs)
        self._ghost_points = ghost_points
        self._internal_points = internal_points
        self.mpi_comm = mpi_comm

    def ghost_points(self, dimension, direction):
        "See ABCBoundary.ghost_points() for documentation." ""
        return self._ghost_points[dimension][self._direction_to_index(direction)]

    def source_and_dest(self, dimension, direction):
        """A wrapper for the MPI_Cart_Shift function via mpi4py.

        To be used in conjuction with a wrapper for MPI_Sendrecv.

        Parameters
        ==========
        dimension: int
            The dimension of the shift
        direction: int
            The direction of the shift

        Returns
        =======
        tuple of ints
            A tuple of ints, ``(source, destination)``, where
            ``source`` is the rank of the mpi node that is the
            source of the data to be sent
            and ``destination`` is the rank of the mpi node that is to
            receive the data.
        """
        return self.mpi_comm.Shift(dimension, direction)

    def internal_slice(self, shape, dimension, direction):
        """See ABCBoundart.internal_slice() for documentation.

        The first tuple in the return of the function is the list of
        slices that identifies the data to be sent. The second tuple
        is the list of slices that identifies the data to be received.

        Note that the return variable for this method is inconsistent
        with the ABCBoundary method. The ABCBoundary method should be
        updated.

        Returns
        =======
        A 2-tuple of lists of tuples of ints
        """

        source, dest = self.source_and_dest(dimension, direction)

        if dest < 0:
            rsend_slice = None
        else:
            send_slice = self._empty_slice(len(shape))
        if source < 0:
            rrecv_slice = None
        else:
            recv_slice = self._empty_slice(len(shape))

        i_point = self._internal_points[dimension][self._direction_to_index(direction)]
        g_point = self._ghost_points[dimension][self._direction_to_index(direction)]
        total_g_points = sum(self._ghost_points[dimension])

        # +1 because the first dimension of the data array is reserved for vector
        # valued data. This may not have been a good desicion
        #
        # The details of why the start and end points for the slices are as they
        # are is complicated. Double check things before you change anything.
        dim_index = dimension + 1

        # This functions are used to to avoid difficulties resulting from the
        # difference between slice(-1, 0) and slice(-1, None), we don't need one
        # for "pos_or_none" as slice(0, x) and slice(None, x) behave the same
        def neg_or_none(number):
            return number if number < 0 else None

        if direction == 1:
            if not dest < 0:
                send_slice[dim_index] = slice(
                    -total_g_points, neg_or_none(-total_g_points + i_point), None
                )
                rsend_slice = tuple(send_slice)
            if not source < 0:
                recv_slice[dim_index] = slice(None, i_point, None)
                rrecv_slice = tuple(recv_slice)
        elif direction == -1:
            if not dest < 0:
                send_slice[dim_index] = slice(
                    max(0, total_g_points - i_point), total_g_points, None
                )
                rsend_slice = tuple(send_slice)
            if not source < 0:
                recv_slice[dim_index] = slice(-i_point, None, None)
                rrecv_slice = tuple(recv_slice)
        return rsend_slice, rrecv_slice

    def external_slice(self, shape, dimension, direction):
        """See ABCBoundary.external_slice() for documentation."""
        if self.mpi_comm.periods[dimension]:
            return self._empty_slice(len(shape))
        coords = self.mpi_comm.Get_coords(self.mpi_comm.rank)
        dim = self.mpi_comm.dims[dimension]
        if coords[dimension] == 0 and direction == -1:
            return super(MPIBoundary, self).external_slice(shape, dimension, -1)
        if coords[dimension] == dim - 1 and direction == 1:
            return super(MPIBoundary, self).external_slice(shape, dimension, 1)
        return self._empty_slice(len(shape))


class SimpleMPIBoundary(MPIBoundary):
    """SimpleMPIBoundary is for boundaries in mpi contexts with a fixed number of
    ghost points in every direction and dimension and the "internal points" equal
    to the ghost points."""

    def __init__(self, ghost_points, *args, **kwargs):
        """The initialiser for SimpleMPIBoundary.

        This class wraps MPIBoundary. It is essentially a short cut for the
        common situation where the number of ghost points is the same for
        every direction and dimension and the number of internal points is
        the same as the number of ghost points.

        Parameters
        ==========
        ghost_points: int
            The number of ghost points.

        """
        number_of_dimensions = kwargs["number_of_dimensions"]
        points_tuple = [
            (ghost_points, ghost_points) for d in range(number_of_dimensions)
        ]
        super(SimpleMPIBoundary, self).__init__(
            points_tuple, points_tuple, *args, **kwargs
        )


class GeneralBoundary(ABCBoundary):
    """GeneralBoundary implements ABCBoundary for grids in an mpi context with arbitrary
    ghost points and arbitrary interal and external slices fixed at run time.

    The ghost points variable must be an array or tuple with one entry for each
    dimension. Each entry is a 2-tuple giving the number of ghost points in the
    negative and positive directions for that axis.

    The data slices variable is an array with one entry for each dimension. Each
    entry is a tuple of slices which when applied to the data stored in a tslice
    gives the data to be communicated / received.
    """

    def __init__(self, ghost_points, internal_slices, external_slices):
        """Inialiser for GeneralBoundary.

        Parameters
        ==========
        ghost_points : list of two tuples of ints
            Each entry of the list gives the number of ghost points for each direction
            for each dimension.
        internal_slices: list of tuples of slices
            Slices which identify the internal boundaries.
        external_slices: list of tuples of slices
            Slices which identify the external boundaries.
        """
        self._ghost_points = ghost_points
        self._interal_slices = internal_slices
        self._external_slices = external_slices

    def ghost_points(self, dimension, direction):
        """See ABCBoundary.ghost_points() for documentation."""
        return self._ghost_points[dimension][self._direction_to_index(direction)]

    def external_slices(self, shape, dimension, direction):
        """See ABCBoundary.external_slices() for documentation."""
        return self._external_slices[dimension][self._direction_to_index(direction)]


################################################################################
# Base Grid class
################################################################################
class ABCGrid(object, metaclass=abc.ABCMeta):
    """The abstract base class for Grid objects.

    A grid class understands both the total computational domain as well as the
    domain associated to the particular MPI node that computation is performed
    on. Each grid object wraps a boundary object and abstracts the boundary
    object api.

    """

    def __init__(
        self,
        shape,
        bounds,
        name="Grid",
        comparison=None,
        mpi=None,
        boundary_data=None,
        *args,
        **kwds,
    ):
        """The initialiser for grid objects.

        Parameters
        ==========
        shape: tuple of ints,
            The shape of a theoretical numpy array that represents the computational
            domain.
        bounds: list of tuples of floats
            For each dimension the tuple gives the maximum and minimum extent of the
            global computational domain
        name: string, Optional
            A name for this grid. Used when logging.
        comparison: Optional
            A argument used when performing comparison of different simulations
            on different grids. Stored in the hdf file during computation. It
            allows for easy identification of how different computation grids
            should be compared.
        mpi: mpi.MPIInterface, Optional
            An instance of an mpi4py wrapped MPI_COMM object.
        boundary_data: ABCBoundary, Optional
            An implementation of ABCBoundary
        """
        self.mpi = mpi
        self.dim = len(shape)
        self.name = name
        self.comparison = comparison
        self.shape = shape
        self.bounds = bounds
        self.boundary_data = boundary_data

    def __strs__(self):
        return self.name

    def __repr__(self):
        return "<%s shape=%s, bounds=%s, comparison=%s, mpi=%s>" % (
            self.name,
            self.shape,
            self.bounds,
            self.comparison,
            self.mpi,
        )

    def communicate(self, data, ghost_point_processor=None):
        """Wraps the mpi.mpiinterface.communicate() method.

        The ghost_point_processor is an arbitary function that
        takes the data and the returned information of the
        mpi.mpiinterfaces.communicate() method, called ``b_values`` here.
        It must modify
        these data structure in place. It is called directly
        before the now modified data and ``b_values`` is returned.

        Parameters
        ==========
        data:
            The data to be communicated
        ghost_point_processor: function(data, grid.ABCBoundary)

        Returns
        =======
        tuple:
            The first element of the tuple is the data that was passed
            to this method. The second element is the returned information
            to the mpi.mpiinterfaces.communicate() method.
        """
        if self.mpi is None:
            return data
        b_values = self.mpi.communicate(data, self.boundary_data)
        if ghost_point_processor:
            ghost_point_processor(data, b_values)
        return data, b_values

    def barrier(self):
        """Provides an easy way to call the MPI_Barrier() method via
        the mpi.mpiinterface object."""
        if self.mpi is None:
            return
        return self.mpi.barrier()

    def external_slices(self, data_shape):
        """Wraps ABCBoundary.external_slices.

        See ABCBoundary.external_slices() for documentation.
        """
        return self.boundary_data.external_slices(data_shape)

    def collect_data(self, data):
        """Wraps mpi.mpiinterfaces.collect_data().

        See mpi.mpiinterfaces.collect_data() for documentation."""
        if self.mpi is None:
            return data
        return self.mpi.collect_data(data)

    @abc.abstractproperty
    def axes(self):
        pass

    @abc.abstractproperty
    def step_sizes(self):
        pass

    @property
    def meshes(self):
        """Generates numpy arrays of grid points.

        The method uses be.lib.stride_tricks.as_strided to cut down on computational
        time. It relies on information in the object itself to handle shape,
        number of points and values.

        The meshes are useful for vectorised functions when computing things like
        initial data for the system.

        Returns
        =======
        list of numpy.ndarray:
            A list of the coordinate values by dimension.
        """
        axes = self.axes
        grid_shape = tuple([axis.size for axis in axes])
        mesh = []
        for i, axis in enumerate(axes):
            strides = be.zeros((len(self.shape),), dtype=int)
            strides[i] = axis.itemsize
            mesh += [
                be.lib_stride_tricks_as_strided(
                    axis, grid_shape, strides, writeable=False
                )
            ]
        return mesh


################################################################################
# Constructors for specific cases
################################################################################
class UniformCart(ABCGrid):
    """An implementation of ABCGrid that assumes that the grid has uniform
    steps and is Cartesian."""

    def __init__(
        self, shape, bounds, mpi_comm=None, comparison=None, name=None, *args, **kwds
    ):
        """Initialiser for UniformCart.

        Parameters
        ==========
        shape: tuple of ints,
            The shape of a theoretical numpy array that represents the computational
            domain.
        bounds: list of tuples of floats
            For each dimension the tuple gives the maximum and minimum extent of the
            global computational domain
        mpi_comm: mpi.MPIInterface, Optional
            An instance of an mpi4py wrapped MPI_COMM object.
        comparison: Optional
            A argument used when performing comparison of different simulations
            on different grids. Stored in the hdf file during computation. It
            allows for easy identification of how different computation grids
            should be compared.
        name: string
            The name of this object. Used in output to hdf files and logging.
        """

        _shape = tuple([s + 1 for s in shape])
        if mpi_comm is None:
            mpi = None
        else:
            mpi = mpiinterfaces.EvenCart(
                _shape,
                kwds.get("boundary_data", None),
                mpi_comm=mpi_comm,
            )
        if name is None:
            name = "UniformCart%s%s%s" % (shape, bounds, comparison)
        super(UniformCart, self).__init__(
            shape, bounds, name=name, comparison=comparison, mpi=mpi, *args, **kwds
        )
        _axes = [
            be.linspace(self.bounds[i][0], self.bounds[i][1], self.shape[i] + 1)
            for i in range(len(self.bounds))
        ]
        self._step_sizes = [axis[1] - axis[0] for axis in _axes]
        if self.mpi is None:
            self._axes = _axes
        else:
            self._axes = [axis[self.mpi.subdomain[i]] for i, axis in enumerate(_axes)]

    @property
    def axes(self):
        """Returns the coordinate values over each dimension.

        Returns
        =======
        numpy.ndarray
        """
        return self._axes

    @property
    def full_grid(self):
        """Returns a UniformCart object that represents the total
        computational domain, rather than just the domain used by this MPI
        node.

        Returns
        =======
        UniformCart
        """
        if self.mpi is None:
            return self
        return UniformCart(
            self.shape, self.bounds, comparison=self.comparison, name=self.name
        )

    @property
    def step_sizes(self):
        """Returns a numpy array containing the fixed step size in each dimension.

        Returns
        =======
        numpy.ndarray
        """
        return self._step_sizes


class GeneralGrid(ABCGrid):
    """An implementation of ABCGrid that allows for periodic dimensions.

    It currently does not allow for use in conjunction with MPI. Note that
    UniformCart and a period MPI network topology are compatible, however.
    """

    def __init__(
        self, shape, bounds, periods, comparison=None, name=None, *args, **kwds
    ):
        """The initialiser for GeneralGrid.

        Parameters
        ==========
        shape: tuple of ints,
            The shape of a theoretical numpy array that represents the computational
            domain.
        bounds: list of tuples of floats
            For each dimension the tuple gives the maximum and minimum extent of the
            global computational domain
        periods: list of ints
            A non-zero value is assumed to indicated that the dimension at the
            given index is periodic. This is modelled on MPI periods.
        comparison: Optional
            A argument used when performing comparison of different simulations
            on different grids. Stored in the hdf file during computation. It
            allows for easy identification of how different computation grids
            should be compared.
        name: string
            The name of this object. Used in output to hdf files and logging.
        """
        _shape = []
        for i, p in enumerate(periods):
            if p:
                _shape.append(shape[i])
            else:
                shape.append(shape[i] + 1)
        _shape = tuple(_shape)
        mpi = None
        if name is None:
            name = "<GeneralGrid shape=%s, bounds=%s,periods=%s, comparison=%s>" % (
                shape,
                bounds,
                periods,
                comparison,
            )
        super(GeneralGrid, self).__init__(
            shape, bounds, name=name, comparison=comparison, mpi=mpi, *args, **kwds
        )
        _axes = [
            be.linspace(self.bounds[i][0], self.bounds[i][1], self.shape[i] + 1)
            for i in range(len(self.bounds))
        ]
        self._step_sizes = [axis[1] - axis[0] for axis in _axes]
        # self._axes = [axis[self.mpi.subdomain] for axis in _axes]

    @property
    def axes(self):
        """Returns the coordinate values over each dimension.

        Returns
        =======
        numpy.ndarray
        """
        return self._axes

    @property
    def full_grid(self):
        """Returns a GeneralGrid object that represents the total
        computational domain, rather than just the domain used by this MPI
        node.

        Of course, since GeneralGrid's don't use MPI, this method returns self.
        It is included for compatibility.

        Returns
        =======
        GeneralGrid
        """
        return self

    @property
    def step_sizes(self):
        """Returns a numpy array containing the fixed step size in each dimension.

        Returns
        =======
        numpy.ndarray
        """
        return self._step_sizes



"""This module writes the data of a timeslice to an hdf file.

The main class SimOutput uses objects of class SimOutput.SimOutputType to
actually write the data. By subclassing SimOutput.SimOutputType in the manner
specified below non - numpy array data can be stored conviently.

I think of this method as a plugin for the system. The user writes a new output
method for some kind of new data type in a sub class of SimOutputType and
'plugs' this method into the 'action' list of SimOutput.
"""
from builtins import str
from builtins import object
from coffee.settings import be

from .actions import Prototype
from ..io.simulation_data import dgTypes, systemD, DataGroup, sysDTypes
from functools import reduce


# A utility class to write to SimulationHDF via the
# actions.py framework.
class SimOutput(Prototype):
    """An action to handle output of data to hdf.

    Each piece of data to be output is passed, as an object, to this class in
    the array actionTypes.

    These objects are subclasses of SimOutputType a class which is accessible
    as an attribute of SimOutput.

    Note that there is a dependence on io.simulation data. In particular
    on the dgTypes dictionary which provides information on the appropriate
    names for created datagroups.
    For information about how to access this data please see the simulation_data
    module.

    Details on how to subclass SimOutputType are given in the documentation for
    the SimOutputType class.
    """

    def __init__(
        self,
        hdf_file,
        solver,
        theSystem,
        theInterval,
        actionTypes,
        cmp_=None,
        overwrite=True,
        name=None,
        **kwds,
    ):
        """The constructor for the SimOutput action.

        The goal for this object is to allow for complete replication of the
        simulation by storing the needed information in the hdf file. This
        requires that the system and solver being used have sensible string
        representations that allow for rebuilding exact duplicates of the
        corresponding objects used in the original simulation.

        Needless to say that this is not currently the case, but I think its a
        good idea.

        Parameters
        ==========
        hdf_file : h5py.File
            This is the file to which data will be written.
        solver : solvers.Solver
        theSystem : systems.System
        theInterval : grid.ABCGrid
        actiontypes : list of SimOutputType objects
                      Each time the SimOutput
                      action is run this list is iterated through and each
                      SimOutputType object is called. This object writes out
                      data according to it's own routine and stores it in the
                      given hdf file.
        cmp_ :
               Something that allows for comparison of simulations. In one
               dimensional simulations this is the number of grid points. This
               is used by the error script to work out which simulations should
               be compared.

        overwrite : bool
                    If you are outputting a simulation into an existing hdf
                    file do you want to allow the simulation to overwrite the
                    existing data?
        name : string
            What is the name of this simulation?
        """
        super(SimOutput, self).__init__(**kwds)
        if name == None:
            hour = str(time.localtime()[3])
            minute = str(time.localtime()[4])
            second = str(time.localtime()[5])
            name = hour + ":" + minute + ":" + second
        self.name = name
        self.hdf = hdf_file
        self.solver = solver
        self.system = theSystem
        self.domain = theInterval
        self.actions = actionTypes
        self.overwrite = overwrite
        self.cmp_ = cmp_
        for action in actionTypes:
            action.setup(self)

    def _doit(self, it, u):
        pass

    class SimOutputType(object):
        """If you want to customise the output to the hdf file subclass this
        class.

        This class handles the organisation of the hdf file structure so that
        you don't have to. It slices, it dices, it picks the group to store the
        datasets that will contain your data. That data group is stored in the
        atribute self.data_group.

        It also handles the inclusion of derived attribute in the .attrs
        variable of the relevant data_set. Sometimes this is needed if more
        than a numpy array needs to be retreived.

        """

        def __init__(self, derivedAttrs=None):
            """Initialiser for SimOutputType

            Parameters
            ==========
            derivedAttrs : A dictionary of keywords and functions, Optional
               The functions must have signature function(iteration,
               tslice, system) and must return an object that h5py
               knows how to store in an hdf file. That object is
               stored in self.data_group[it].attrs[key].
            """
            if derivedAttrs is None:
                self.derivedAttrs = {}
            else:
                self.derivedAttrs = derivedAttrs

        def setup(self, parent):
            """This method is called by SimOuput on construction.

            Overide this is additional setup as needed, in particular if data
            should be extracted from parent. See SimOutputType.System
            for an example.

            This method should always to called so if you do override it, make
            sure to include a call to super(YourClass, self).setup(partent).
            This is necessary because without the call the data_group may not
            be configured correctly.

            Parameters
            ==========
            SimOuput:
                parent the SimOutput class that this SimOutputType belongs to.
            """
            if parent.overwrite:
                self.data_group = DataGroup(
                    parent.hdf.require_group(self.groupname).require_group(parent.name)
                )
            else:
                self.data_group = DataGroup(
                    parent.hdf.create_group(self.groupname).create_group(parent.name)
                )
            self.parent = parent
            self.data_group.attrs["cmp"] = parent.cmp_

        def __call__(self, it, u):
            for key, value in list(self.derivedAttrs.items()):
                v = value(it, u, self.parent.system)
                self.data_group[it].attrs[key] = v

    class Data(SimOutputType):
        """Writes out tslices.TimeSlice.data.

        This class assumes that tslice.TimeSlice.data has a type that h5py
        understands.
        """

        groupname = dgTypes["raw"]

        def __call__(self, it, u):
            dg = self.data_group
            dg[it] = u.data
            super(SimOutput.Data, self).__call__(it, u)

    class Exact(SimOutputType):
        """Calls system.exact_value to write out the exact value of the system.

        This is useful if, during error calculation, you want the error rates
        against an exact solution.

        """

        groupname = dgTypes["exact"]

        def __call__(self, it, u):
            dg = self.data_group
            parent = self.parent
            dg[it] = parent.system.exact_value(u.time, u.domain).data
            super(SimOutput.Exact, self).__call__(it, u)

    class Times(SimOutputType):
        """Would you like to know what time the data at a particular data_set
        is meant to correspond to?

        Then you need this SimOutputType.

        """

        groupname = dgTypes["time"]

        def __call__(self, it, u):
            dg = self.data_group
            dg[it] = be.array([u.time])
            super(SimOutput.Times, self).__call__(it, u)

    class TimeStep(SimOutputType):
        """Just like time buts tells you what the calculated dt was.

        This can be calculated from the data in Times, only if you are
        recording the data in the next timestep...

        """

        groupname = dgTypes["dt"]

        def __call__(self, it, u):
            dg = self.data_group
            dg[it] = be.array([self.parent.system.timestep(u)])
            super(SimOutput.TimeStep, self).__call__(it, u)

    class Domains(SimOutputType):
        """Records domains of the simulation.

        This SimOutputType does not record the grid object, but rather the axes
        and comparison variables of the grid. Hence no mpi information is
        collected here.

        Currently this is because ibvp.IBVP before it runs actions collates all
        the data together so that, from the point of view of the hdf file there
        is only one process accessing it.

        This will need to be changed at some point. h5py has routines to allow
        more than one process to write to a file. But coffee does not
        currently capitalise on these routines.

        """

        groupname = dgTypes["domain"]

        def __call__(self, it, u):
            dg = self.data_group
            axes = u.domain.axes
            axes_shape = tuple([axis.size for axis in axes])
            axes_flat = be.empty(
                (reduce(lambda x, y: x + y, axes_shape),), dtype=u.domain.axes[0].dtype
            )
            start = 0
            for i, axis in enumerate(axes):
                axes_flat[start : start + axes_shape[i]] = axis
                start = start + axes_shape[i]
            dg[it] = axes_flat
            dg[it].attrs["axes_shape"] = axes_shape
            dg[it].attrs["shape"] = u.domain.shape
            dg[it].attrs["bounds"] = u.domain.bounds
            dg[it].attrs["comparison"] = u.domain.comparison
            super(SimOutput.Domains, self).__call__(it, u)

    class DerivedData(SimOutputType):
        """Runs the data through a user defined function before writing out to
        the data_set.

        The function must have the signature function(int,
        tslice.TimeSlice, system.System) and must return
        data of a type that h5py recognises.

        Note that there is a size limitation on the .attrs variable and
        therefore this SimOutputType maybe more appropriate.
        """

        def __init__(self, name, function, frequency=1, start=0, derivedAttrs=None):
            """The initialiser for DerivedData.

            Parameters
            ==========
            name : string
                The name for the hdf.datagroup.
            function : function(int, tslice.TimeSlice, system)
                The function that returns the data to be stored in
                the datagroup.
            frequency : int
                It dicates how many iterations pass for each call of this action.
            start : float
                The time from which to start performing the action.
            derivedAttrs :
                See the documentation for SimOutputType.
            """
            self.func = function
            self.groupname = name
            self.freq = frequency
            self.start = start
            super(SimOutput.DerivedData, self).__init__(derivedAttrs)

        def __call__(self, it, u):
            if it % self.freq == 0 and self.start < u.time:
                dg = self.data_group
                dg[it] = self.func(it, u, self.parent.system)
                super(SimOutput.DerivedData, self).__call__(it, u)

    class System(SimOutputType):
        """Attempts to write out enough information about the system to allow
        for exact reconstruction of the simulation that produced the data being
        stored.

        Of course most things arn't that simple... and this SimOutput type
        requires care during use.
        """

        groupname = systemD

        def setup(self, parent):
            super(SimOutput.System, self).setup(parent)
            g = self.data_group.group
            psystem = be.asarray(repr(parent.system)).astype("S")
            pgrid = be.asarray(repr(parent.domain)).astype("S")
            psolver = be.asarray(repr(parent.solver)).astype("S")
            pcmp = be.asarray(repr(parent.cmp_)).astype("S")
            pnumvar = be.asarray(repr(parent.system.numvar)).astype("S")

            g.require_dataset(
                sysDTypes["system"], psystem.shape, psystem.dtype, data=psystem
            )
            g.require_dataset(sysDTypes["grid"], pgrid.shape, pgrid.dtype, data=pgrid)
            g.require_dataset(
                sysDTypes["solver"], psolver.shape, psolver.dtype, data=psolver
            )
            g.require_dataset(sysDTypes["cmp"], pcmp.shape, pcmp.dtype, data=pcmp)
            g.require_dataset(
                sysDTypes["numvar"], pnumvar.shape, pnumvar.dtype, data=pnumvar
            )

        def __call__(self, it, u):
            pass



#!/usr/bin/env python
# encoding: utf-8
"""
A settings file that is initialised by the user to configure the backend class.

Should be easily generalised to other settings that have similar requirements.
"""

def init(backend_obj):
    global be
    be = backend_obj


#!/usr/bin/env python
# encoding: utf-8


"""
OneDAdvection_setup.py

Created by Chris Stevens 2023
"""

# Import python libraries
from builtins import range
import sys
import os
import numpy as np
import h5py
import argparse
from mpi4py import MPI

# Import standard code base and initialize NumPy backend
from coffee.settings import init
from coffee.backend import backend as be
my_backend = be.set_backend("numpy")
init(my_backend)

from coffee import ibvp, actions, solvers, grid
from coffee.diffop import fd

# Import system to use
import OneDAdvection
import OneDAdvection_plotter as plotter

np.set_printoptions(threshold=np.inf, precision=16)

################################################################################
# Parser settings 
################################################################################

# Initialise parser
parser = argparse.ArgumentParser(description=\
"""This program numerically solves an IBVP for a simple advection equation.""")

# Parse files
parser.add_argument('-f','-file', help=\
"""The name of the hdf file to be produced.""")
parser.add_argument('-d','-display', default = False, 
    action='store_true', help=\
"""A flag to indicate if visual display is required.""")
args = parser.parse_args()

################################################################################  
# These are the commonly altered settings
################################################################################

# Output settings
store_output = args.f is not None
display_output = args.d
if store_output and args.f is None:
    print("OneDAdvection_setup.py: error: argument -f/-file is required")
    sys.exit(1)

# File settings
if store_output:
    args.f = os.path.abspath(args.f)
    try:
        if not os.path.exists(os.path.split(args.f)[0]):
            os.makedirs(os.path.split(args.f)[0])
    except OSError as oserror:
        if oserror.errno is not errno.EEXIST:
            raise oserror

# How many systems?
num_of_grids = 1

# How many grid points?
N = 200

# What grid to use?
xstart = -1
xstop  = 1

# Times to run between
tstart = 0.0
tstop  = 10.

# Configuration of System
CFL = 0.5
tau = 1.0

# Differential operator
# diffop = fd.FD12()
diffop = fd.FD14()

################################################################################      
## MPI set up                                                                         
################################################################################ 

dims     = MPI.Compute_dims(MPI.COMM_WORLD.size, [0])                                    
periods  = [0]                                                                        
reorder  = True                                                                       
mpi_comm = MPI.COMM_WORLD.Create_cart(dims, periods=periods, reorder=reorder)        

################################################################################
# Grid construction
################################################################################

# Grid point data      
res = [N*2**i for i in range(num_of_grids)]

# Determine the boundary data
ghost_points = (diffop.ghost_points(),)
internal_points = (diffop.internal_points(),)
b_data = grid.MPIBoundary(
    ghost_points, 
    internal_points, 
    mpi_comm=mpi_comm, 
    number_of_dimensions=1
)

# Build grids
grids = [
    grid.UniformCart(
        (res[i],), 
        [[xstart,xstop]],
        comparison = res[i],
        mpi_comm = mpi_comm,
        boundary_data=b_data
    ) 
    for i in range(num_of_grids)
]

################################################################################
# Initialise systems
################################################################################

char_speed = 1. # Must be positive
systems = []
for i in range(num_of_grids):
    systems += [OneDAdvection.OneDAdvection(\
        diffop, tau, CFL, char_speed
        )]
    
# Configuration of IBVP
solvers = [solvers.RungeKutta4(sys) for sys in systems]
maxIteration = 10000000

################################################################################
# Set up hdf file to store output
################################################################################

if store_output and mpi_comm.rank == 0:
    hdf_file = h5py.File(args.f,"w")

################################################################################
# Set up action types for data storage in hdf file
################################################################################

output_actions = [
    actions.SimOutput.Data(),
    actions.SimOutput.Times(),
    actions.SimOutput.System(),
    actions.SimOutput.Domains()
    ]

################################################################################
# Perform computation
################################################################################

for i, system in enumerate(systems):
        
        #Construct Actions
        actionList = []
        if display_output and mpi_comm.rank == 0:
            actionList += [
                plotter.Plotter(
                    system,
                    frequency=1,
                    xlim=(xstart, xstop),
                    ylim=(-1, 1),
                    delay=0.0001
                )
            ]
        if store_output and mpi_comm.rank == 0:
            actionList += [actions.SimOutput(\
                hdf_file,\
                solvers[i], \
                system, \
                grids[i], \
                output_actions,\
                overwrite = True,\
                name = grids[i].name,\
                cmp_ = grids[i].comparison\
                )]
            
        # Construct and run problem
        problem = ibvp.IBVP(solvers[i], system, grid = grids[i],\
                maxIteration = 1000000, action = actionList,\
                minTimestep = 1e-12)
        problem.run(tstart, tstop)


#!/usr/bin/env python
# encoding: utf-8


"""
OneDAdvection.py

Created by Chris Stevens 2023
"""

# Import Python libraries
from coffee.settings import be
from coffee.tslices import tslices
from coffee.system import System

# A simple bump function for the boundary condition
def bump(t):
	return be.sin(8.*t)**7.

# Class that describes a simple advection equation
class OneDAdvection(System):

	############################################################################
	# Constructor
	############################################################################

	def __init__(self, D, tau, CFL, char_speed):
		super(OneDAdvection, self).__init__()
		self.D          = D
		self.tau        = tau
		self.CFL        = CFL
		self.char_speed = char_speed
		self.name       = "1D Advection equation"
		self.numvar     = 1

	############################################################################
	# Boundary functions
	############################################################################
		
	# Left boundary condition
	def left(self, t):
		return bump(t)
	
	############################################################################
	# Evolution Routine
	############################################################################

	def evaluate(self, t, U, intStep = None):

		# Define useful variables
		Psi = U.data[0]
		dx  = U.domain.step_sizes[0]

		# Calculate spatial derivative
		DxPsi = self.D(Psi, dx)

		# Evaluate right hand side
		DtPsi = -self.char_speed*DxPsi

		# Impose boundary condition with SAT method
		pt_l     = self.D.penalty_boundary(dx, "left")
		C_left   = self.tau * pt_l * (-self.char_speed)
		l_Psi    = self.left(t)
		DtPsi[0] -= C_left * (Psi[0] - l_Psi)

		# Output time if desired
		# print('t =', t)

		# Return RHS of evolution equation with SAT boundary condition imposed
		return tslices.TimeSlice([DtPsi], U.domain, time = t)

	############################################################################
	# Initial Data Routine
	############################################################################
	
	def initial_data(self, t0, grid):

		# Set initial data
		x   = grid.meshes[0]
		Psi = be.zeros_like(x)

		# Return Timeslice object
		return tslices.TimeSlice([Psi], grid, time = t0)

	############################################################################
	# Timestep Routine
	############################################################################

	def timestep(self, U):
		return self.CFL * U.domain.step_sizes[0] / self.char_speed


#!/usr/bin/env python
# encoding: utf-8


"""
TwoDAdvection.py

Created by Chris Stevens 2023
"""

# Import Python libraries
from functools import partial

from coffee.settings import be
from coffee.tslices import tslices
from coffee.system import System
from coffee.diffop.fd import ghost_point_processor

# A simple bump function for the boundary condition
def bump(t, grid, grid_global):
	start = grid_global[0]
	stop  = grid_global[-1]
	return be.sin(be.pi*start / (start - stop) - \
	       be.pi / (start - stop)*grid)**8.* \
		   be.sin(8.*t)**7.

# Class that describes a simple advection equation
class TwoDAdvection(System):

	################################
	# Constructor
	################################

	def __init__(self, Dx, Dy, tau, CFL, char_speed_x, char_speed_y, \
	      		x_global, y_global):
		super(TwoDAdvection, self).__init__()
		self.Dx           = Dx
		self.Dy           = Dy
		self.tau          = tau
		self.CFL          = CFL
		self.char_speed_x = char_speed_x
		self.char_speed_y = char_speed_y
		self.x_global     = x_global
		self.y_global     = y_global
		self.name         = "2D Advection equation"
		self.numvar       = 1

	############################################################################
	# Boundary functions
	############################################################################
		
	# Left x boundary condition
	def left_x(self, t, y, y_global):
		return bump(t, y, y_global)
	
	# Left y boundary condition
	def left_y(self, t, x, x_global):
		return bump(t, x, x_global)

	############################################################################
	# Evolution Routine
	############################################################################

	def evaluate(self, t, U, intStep = None):

		# Define useful variables
		Psi = U.data[0]
		x   = U.domain.axes[0]
		y   = U.domain.axes[1]
		dx  = U.domain.step_sizes[0]
		dy  = U.domain.step_sizes[1]

		# Calculate spatial derivatives by applying the 1D FD operators
		# over the mesh
		DxPsi = be.apply_along_axis(
			lambda f:self.Dx(f, dx),
			0, 
			Psi
			)
		DyPsi = be.apply_along_axis(
			lambda f:self.Dy(f, dy),
			1, 
			Psi
			)

		# Communicate spatial derivative through neighbouring MPI process(es)
		new_derivatives, _ = U.communicate(
			partial(ghost_point_processor),
			data=be.array([
				DxPsi, DyPsi
			])
		)
		DxPsi, DyPsi = new_derivatives

		# Impose boundary conditions with SAT method
		pt_x_l = self.Dx.penalty_boundary(dx, "left")
		pt_y_l = self.Dy.penalty_boundary(dy, "left")
	
		# Only two of the four boundaries require boundary conditions
		C_x_left     = self.tau * pt_x_l * (-self.char_speed_x)
		l_x_Psi      = self.left_x(t, y, self.y_global)
		C_y_left     = self.tau * pt_y_l * (-self.char_speed_y)
		l_y_Psi      = self.left_y(t, x, self.x_global)

		# Evaluate right hand side
		DtPsi = -self.char_speed_x*DxPsi - self.char_speed_y*DyPsi

		# Do the external boundaries
		b_data = U.external_slices()
		for dim, direction, _ in b_data:
			if dim == 0 and direction == -1: # Left x boundary)
				DtPsi[0,:] -= C_x_left * \
					(Psi[0,:] - l_x_Psi)
			elif dim == 1 and direction == -1: # Left y boundary
				DtPsi[:,0] -= C_y_left * \
					(Psi[:,0] - l_y_Psi)

		# # Output time if desired
		# print('t =', t)

		# Return RHS of evolution equation with SAT boundary condition imposed
		return tslices.TimeSlice([DtPsi], U.domain, time = t)

	############################################################################
	# Initial Data Routine
	############################################################################

	def initial_data(self, t0, grid):

		# Set initial data
		x   = grid.meshes[0]
		Psi = be.zeros_like(x)

		# Return Timeslice object
		return tslices.TimeSlice([Psi], grid, time = t0)

	############################################################################
	# Timestep Routine
	############################################################################

	def timestep(self, U):

		return self.CFL * min(U.domain.step_sizes[0] / self.char_speed_x, \
	  						  U.domain.step_sizes[1] / self.char_speed_y)
#!/usr/bin/env python
# encoding: utf-8


"""
TwoDAdvection.py

Created by Chris Stevens 2023
"""

# Import Python libraries
import numpy as np
from functools import partial

from coffee.tslices import tslices
from coffee.system import System
from coffee.diffop.fd import ghost_point_processor

# A simple bump function for the boundary condition
def bump(t, grid, grid_global):
	start = grid_global[0]
	stop  = grid_global[-1]
	return np.sin(np.pi*start / (start - stop) - \
	       np.pi / (start - stop)*grid)**8.* \
		   np.sin(8.*t)**7.

# Class that describes a simple advection equation
class TwoDAdvection(System):

	################################
	# Constructor
	################################

	def __init__(self, Dx, Dy, tau, CFL, char_speed_x, char_speed_y, \
	      		x_global, y_global):
		super(TwoDAdvection, self).__init__()
		self.Dx           = Dx
		self.Dy           = Dy
		self.tau          = tau
		self.CFL          = CFL
		self.char_speed_x = char_speed_x
		self.char_speed_y = char_speed_y
		self.x_global     = x_global
		self.y_global     = y_global
		self.name         = "2D Advection equation"
		self.numvar       = 1

	############################################################################
	# Boundary functions
	############################################################################
		
	# Left x boundary condition
	def left_x(self, t, y, y_global):
		return bump(t, y, y_global)
	
	# Left y boundary condition
	def left_y(self, t, x, x_global):
		return bump(t, x, x_global)

	############################################################################
	# Evolution Routine
	############################################################################

	def evaluate(self, t, U, intStep = None):

		# Define useful variables
		Psi = U.data[0]
		x   = U.domain.axes[0]
		y   = U.domain.axes[1]
		dx  = U.domain.step_sizes[0]
		dy  = U.domain.step_sizes[1]

		# Calculate spatial derivatives by applying the 1D FD operators
		# over the mesh
		DxPsi = np.apply_along_axis(
			lambda f:self.Dx(f, dx),
			0, 
			Psi
			)
		DyPsi = np.apply_along_axis(
			lambda f:self.Dy(f, dy),
			1, 
			Psi
			)

		# Communicate spatial derivative through neighbouring MPI process(es)
		new_derivatives, _ = U.communicate(
			partial(ghost_point_processor),
			data=np.array([
				DxPsi, DyPsi
			])
		)
		DxPsi, DyPsi = new_derivatives

		# Impose boundary conditions with SAT method
		pt_x_l = self.Dx.penalty_boundary(dx, "left")
		pt_y_l = self.Dy.penalty_boundary(dy, "left")
	
		# Only two of the four boundaries require boundary conditions
		C_x_left     = self.tau * pt_x_l * (-self.char_speed_x)
		l_x_Psi      = self.left_x(t, y, self.y_global)
		C_y_left     = self.tau * pt_y_l * (-self.char_speed_y)
		l_y_Psi      = self.left_y(t, x, self.x_global)

		# Evaluate right hand side
		DtPsi = -self.char_speed_x*DxPsi - self.char_speed_y*DyPsi

		# Do the external boundaries
		b_data = U.external_slices()
		for dim, direction, _ in b_data:
			if dim == 0 and direction == -1: # Left x boundary)
				DtPsi[0,:] -= C_x_left * \
					(Psi[0,:] - l_x_Psi)
			elif dim == 1 and direction == -1: # Left y boundary
				DtPsi[:,0] -= C_y_left * \
					(Psi[:,0] - l_y_Psi)

		# # Output time if desired
		# print('t =', t)

		# Return RHS of evolution equation with SAT boundary condition imposed
		return tslices.TimeSlice([DtPsi], U.domain, time = t)

	############################################################################
	# Initial Data Routine
	############################################################################

	def initial_data(self, t0, grid):

		# Set initial data
		x   = grid.meshes[0]
		Psi = np.zeros_like(x)

		# Return Timeslice object
		return tslices.TimeSlice([Psi], grid, time = t0)

	############################################################################
	# Timestep Routine
	############################################################################

	def timestep(self, U):

		return self.CFL * min(U.domain.step_sizes[0] / self.char_speed_x, \
	  						  U.domain.step_sizes[1] / self.char_speed_y)


import argparse
import sys
from builtins import range

import backend as be

from coffee.io import simulation_data as sd

be.set_printoptions(threshold=be.inf)

parser = argparse.ArgumentParser(
    description="""This program compares the contents of two hdf files as produced by coffee."""
)

parser.add_argument("-f1", help="""The first file to be compared.""")
parser.add_argument("-f2", help="""The second file to be compared.""")
parser.add_argument(
    "-dg",
    "-dgtype",
    help="""A data group to be analysed. See simulation_data for information on
the different data groups. Defaults to "raw". Multiple data groups can be given.""",
)

args = parser.parse_args()
if args.dg is None:
    args.dg = "raw"

first = sd.SimulationHDF(args.f1)
second = sd.SimulationHDF(args.f2)

first_sims = first.getSims()
second_sims = second.getSims()

if first_sims != second_sims:
    print("Different simulations in files")
    sys.exit()


def get_sim(sims, name):
    for sim in sims:
        if sim.name == name:
            return sim
    raise ValueError("No sim named " + name)


for f_sim in first_sims:
    s_sim = get_sim(second_sims, f_sim.name)
    f_datagroup = f_sim.getDgType(args.dg)
    s_datagroup = s_sim.getDgType(args.dg)

    if len(f_datagroup) != len(s_datagroup):
        print(
            "The lengths of %s datagroup in simulations %s and %s do not match"
            % (args.dg, f_sim.name, s_sim.name)
        )
        sys.exit()

    index = 0
    while index < len(f_datagroup):
        f_data = f_datagroup[index][()]
        s_data = s_datagroup[index][()]
        if not be.array_equal(f_data, s_data):
            for i in range(len(f_data)):
                if not be.array_equal(f_data[i], s_data[i]):
                    print("Difference in component %s" % repr(i))
                    for j in range(len(f_data[i])):
                        if f_data[i][j] != s_data[i][j]:
                            print("Difference in element %s" % repr(j))
                            print("%s != %s" % (repr(f_data[i][j]), repr(s_data[i][j])))
                    print(
                        "Different data encountered at index %i:\n%s\n%s"
                        % (index, repr(f_data[i]), repr(s_data[i]))
                    )
            sys.exit()
        index += 1

print("DataType %s in files %s and %s is the same" % (args.dg, args.f1, args.f2))



import pytest

from coffee.backend import Backend


class MockBackend:
    def __init__(self):
        self.calls = []

    def __getattr__(self, name):
        def method(*args, **kwargs):
            self.calls.append((name, args, kwargs))
            return f"{name} called"

        return method


class TestBackend:
    @pytest.fixture
    def backend(self):
        return Backend()

    def test_set_backend_default(self, backend):
        backend.set_backend()
        assert backend.backend_name == "numpy"

    def test_set_backend_unsupported(self, backend):
        with pytest.raises(ValueError):
            backend.set_backend("unsupported")

    def test_get_backend_unsupported(self, backend):
        with pytest.raises(ValueError):
            backend._get_backend("unsupported")

    def test_backend_methods(self, backend, monkeypatch):
        mock_backend = MockBackend()
        monkeypatch.setattr(backend, "_backend", mock_backend)

        backend.abs(1)
        backend.array([1, 2, 3])
        backend.linspace(0, 1, 5)
        backend.zeros((3, 3))
        backend.sum([1, 2, 3])

        expected_calls = [
            ("abs", (1,), {}),
            ("array", ([1, 2, 3],), {}),
            ("linspace", (0, 1, 5), {}),
            ("zeros", ((3, 3),), {}),
            ("sum", ([1, 2, 3],), {}),
        ]

        assert mock_backend.calls == expected_calls
